@@ -69,7 +69,8 @@ namespace aspeller {
   Dictionary::Dictionary(BasicType t, const char * n)
     : Cacheable(&dict_cache), lang_(), id_(), 
       basic_type(t), class_name(n),
-      affix_compressed(false), have_soundslike(false), 
+      affix_compressed(false), 
+      invisible_soundslike(false), soundslike_root_only(false),
       fast_scan(false), fast_lookup(false)
   {
     id_.reset(new Id(this));
@@ -177,7 +178,7 @@ namespace aspeller {
 
   PosibErr<void> Dictionary::add_repl(ParmString mis, ParmString cor) 
   {
-    if (have_soundslike) {
+    if (!invisible_soundslike) {
       VARARRAY(char, sl, mis.size() + 1);
       lang()->LangImpl::to_soundslike(sl, mis.str(), mis.size());
       return add_repl(mis, cor, sl);
@@ -188,7 +189,7 @@ namespace aspeller {
 
   PosibErr<void> Dictionary::add(ParmString w) 
   {
-    if (have_soundslike) {
+    if (!invisible_soundslike) {
       VARARRAY(char, sl, w.size() + 1);
       lang()->LangImpl::to_soundslike(sl, w.str(), w.size());
       return add(w, sl);
@@ -322,13 +323,11 @@ namespace aspeller {
 
   OStream & WordEntry::write (OStream & o,
                               const Language & l,
-                              const ConvertWord & cw,
                               Convert * c) const
   {
     String w;
     CharVector buf;
-    cw.convert(word, w);
-    write_conv(w);
+    write_conv(word);
     if (aff && *aff) {
       o << '/';
       write_conv(aff);
@@ -472,20 +471,13 @@ namespace aspeller {
   void LocalDictInfo::set_language(const Language * l)
   {
     compare.lang = l;
-    convert.lang = l;
   }
 
-  void LocalDictInfo::set(const Language * l, const Config & c, bool strip)
+  void LocalDictInfo::set(const Language * l, const Config & c)
   {
-    if (c.have("strip-accents"))
-      strip = c.retrieve_bool("strip-accents");
-
     compare.lang = l;
     compare.case_insensitive = c.retrieve_bool("ignore-case");
     compare.ignore_accents   = c.retrieve_bool("ignore-accents");
-    compare.strip_accents    = strip;
-    convert.lang = l;
-    convert.strip_accents = strip;
   }
   
 }
