@@ -58,11 +58,11 @@ namespace aspeller {
     , {"soundslike",          KeyInfoString, "none", ""}
     , {"special",             KeyInfoString, "", ""}
     , {"ignore-accents" ,     KeyInfoBool, "", "", 0, FOR_CONFIG}
-    , {"use-soundslike" ,     KeyInfoBool, "", "", 0, FOR_CONFIG}
-    , {"use-jump-tables",     KeyInfoBool, "", "", 0, FOR_CONFIG}
+    , {"invisible-soundslike",KeyInfoBool, "", "", 0, FOR_CONFIG}
     , {"keyboard",            KeyInfoString, "standard", "", 0, FOR_CONFIG} 
     , {"affix",               KeyInfoString, "none", ""}
     , {"affix-compress",      KeyInfoBool, "false", "", 0, FOR_CONFIG}
+    , {"partially-expand",    KeyInfoBool, "false", "", 0, FOR_CONFIG}
     , {"affix-char",          KeyInfoString, "/", "", 0, FOR_CONFIG}
     , {"flag-char",           KeyInfoString, ":", "", 0, FOR_CONFIG}
     , {"repl-table",          KeyInfoString, "none", ""}
@@ -187,8 +187,8 @@ namespace aspeller {
       de_accent_[i] = to_plain_[i] == 0 ? to_uchar(i) : to_plain_[i];
     }
 
-    to_plain_[0] = 1; // to make things slightly easier // FIXME: Should this be 0x10
-    to_plain_[1] = 1;
+    to_plain_[0] = 0x10; // to make things slightly easier
+    to_plain_[1] = 0x10;
 
     for (unsigned int i = 0; i != 256; ++i) {
       to_stripped_[i] = to_plain_[(unsigned char)to_lower_[i]];
@@ -213,8 +213,9 @@ namespace aspeller {
       clean_is = to_lower_;
     }
 
-    for (int i = 0; i != 256; ++i) {
-      to_clean_[i] = char_type_[i] >= Letter ? clean_is[i] : 0;
+    for (unsigned i = 0; i != 256; ++i) {
+      to_clean_[i] = char_type_[i] > NonLetter ? clean_is[i] : 0;
+      if ((unsigned char)to_clean_[i] == i) char_info_[i] |= CLEAN;
     }
 
     to_clean_[0x00] = 0x10; // to make things slightly easier
@@ -257,6 +258,8 @@ namespace aspeller {
     soundslike_.reset(pe);
     soundslike_chars_ = soundslike_->soundslike_chars();
 
+    have_soundslike_ = strcmp(soundslike_->name(), "none") != 0;
+
     //
     // prep affix code
     //
@@ -268,6 +271,7 @@ namespace aspeller {
     //
 
     String repl = data.retrieve("repl-table");
+    have_repl_ = false;
     if (repl != "none") {
 
       String repl_file;
@@ -284,6 +288,8 @@ namespace aspeller {
         }
       }
       repls_.resize(num_repl);
+      if (num_repl > 0)
+        have_repl_ = true;
 
       for (size_t i = 0; i != num_repl; ++i) {
         bool res = getdata_pair(REPL, d, buf);
@@ -292,7 +298,9 @@ namespace aspeller {
         assert(d.key == "rep"); // FIXME
         split(d);
         repls_[i].substr = buf_.dup(iconv(d.key));
+        to_clean((char *)repls_[i].substr, repls_[i].substr);
         repls_[i].repl   = buf_.dup(iconv(d.value));
+        to_clean((char *)repls_[i].repl, repls_[i].repl);
       }
 
     }
@@ -321,32 +329,30 @@ namespace aspeller {
 
   WordInfo Language::get_word_info(ParmString str) const
   {
-    CharInfo first, all = CHAR_INFO_ALL;
+    CharInfo first = CHAR_INFO_ALL, all = CHAR_INFO_ALL;
     const char * p = str;
-    while (first = char_info(*p++), all &= first, !(first & LETTER))
-      if (!*p) return AllUpper | ALL_PLAIN;
+    while (*p && (first = char_info(*p++), all &= first, !(first & LETTER)));
     while (*p) all &= char_info(*p++);
     WordInfo res;
-    if      (all & UPPER)   res = AllUpper;
-    else if (all & LOWER)   res = AllLower;
+    if      (all & LOWER)   res = AllLower;
+    else if (all & UPPER)   res = AllUpper;
     else if (first & TITLE) res = FirstUpper;
     else                    res = Other;
-    if (all & LETTER) res |= ALL_LETTER;
     if (all & PLAIN)  res |= ALL_PLAIN;
+    if (all & CLEAN)  res |= ALL_CLEAN;
     return res;
   }
   
   CasePattern Language::case_pattern(ParmString str) const  
   {
-    CharInfo first, all = CHAR_INFO_ALL;
+    CharInfo first = CHAR_INFO_ALL, all = CHAR_INFO_ALL;
     const char * p = str;
-    while (first = char_info(*p++), all &= first, !(first & LETTER))
-      if (!*p) return AllUpper;
+    while (*p && (first = char_info(*p++), all &= first, !(first & LETTER)));
     while (*p) all &= char_info(*p++);
-      if      (all & UPPER)   return AllUpper;
-      else if (all & LOWER)   return AllLower;
-      else if (first & TITLE) return FirstUpper;
-      else                    return Other;
+    if      (all & LOWER)   return AllLower;
+    else if (all & UPPER)   return AllUpper;
+    else if (first & TITLE) return FirstUpper;
+    else                    return Other;
   }
   
   void Language::fix_case(CasePattern case_pattern,
@@ -370,7 +376,7 @@ namespace aspeller {
   }
 
   const char * Language::fix_case(CasePattern case_pattern, const char * str,
-                                  String buf) const 
+                                  String & buf) const 
   {
     if (!str[0]) return str;
     if (case_pattern == AllUpper) {
@@ -386,102 +392,32 @@ namespace aspeller {
       return str;
     }
   }
-    
-  bool SensitiveCompare::operator() (const char * word, 
-				     const char * inlist) const
+  
+  bool SensitiveCompare::operator() (const char * word0, 
+				     const char * inlist0) const
   {
-   // this will fail if word or inlist is empty
-    assert (*word != '\0' && *inlist != '\0');
-    
-    // if begin inlist is a begin char then it must match begin word
-    // chop all begin chars from the begin of word and inlist  
-    if (lang->special(*inlist).begin) {
-      if (*word != *inlist)
-	return false;
-      ++word, ++inlist;
-    } else if (lang->special(*word).begin) {
-      ++word;
-    }
-    
-    // this will fail if word or inlist only contain a begin char
-    assert (*word != '\0' && *inlist != '\0');
-    
-    if (case_insensitive) {
-      if (ignore_accents) {
-
-	while (*word != '\0' && *inlist != '\0') 
-	  ++word, ++inlist;
-
-      } else if (strip_accents) {
-
-	while (*word != '\0' && *inlist != '\0') {
-	  if (lang->to_lower(*word) != lang->to_plain(lang->to_lower(*inlist)))
-	    return false;
-	  ++word, ++inlist;
-	}
-
-      } else {
-
-	while (*word != '\0' && *inlist != '\0') {
-	  if (lang->to_lower(*word) != lang->to_lower(*inlist))
-	    return false;
-	  ++word, ++inlist;
-	}
-
-      }
-    } else {
-      //   (note: there are 3 possible casing lower, upper and title)
-      //   if is lower begin inlist then begin word can be any casing
-      //   if not                   then begin word must be the same case
-      bool case_compatible = true;
-      if (!ignore_accents) {
-	if (strip_accents) {
-	  if (lang->to_lower(*word) != lang->to_plain(lang->to_lower(*inlist)))
-	    return false;
-	} else {
-	  if (lang->to_lower(*word) != lang->to_lower(*inlist))
-	    return false;
-	}
-      }
-      if (!lang->is_lower(*inlist) && lang->to_plain(*word) != lang->to_plain(*inlist))
-	case_compatible = false;
-      bool all_upper = lang->is_upper(*word);
-      ++word, ++inlist;
-      while (*word != '\0' && *inlist != '\0') {
-        if (!lang->is_upper(*word))
-          all_upper = false;
-        if (ignore_accents) {
-          if (lang->to_plain(*word) != lang->to_plain(*inlist))
-            if (lang->to_stripped(*word) != lang->to_stripped(*inlist))
-              return false;
-            else
-              case_compatible = false;
-        } else if (strip_accents) {
-          if (*word != lang->to_plain(*inlist))
-            if (lang->to_lower(*word) != lang->to_plain(lang->to_lower(*inlist)))
-              return false;
-            else // accents match case does not
-              case_compatible = false;
-        } else {
-          if (*word != *inlist)
-            if (lang->to_lower(*word) != lang->to_lower(*inlist))
-              return false;
-            else // accents match case does not
-              case_compatible = false;
-        }
-        ++word, ++inlist;
-      }
-      //   if word is all upper than the casing of inlist can be anything
-      //   otherwise the casing of tail begin and tail inlist must match
-      if (all_upper) 
-	case_compatible = true;
-      if (!case_compatible) 
-	return false;
-    }
-
+    assert(*word0 && *inlist0);
+  try_again:
+    const char * word = word0;
+    const char * inlist = inlist0;
+    if (*word == *inlist || *word == lang->to_title(*inlist)) ++word, ++inlist;
+    else                                                      goto try_upper;
+    while (*word && *inlist && *word == *inlist) ++word, ++inlist;
+    if (*inlist) goto try_upper;
     if (lang->special(*word).end) ++word;
-
-    return (*word == '\0' && *inlist == '\0');
+    if (*word) goto try_upper;
+    return true;
+  try_upper:
+    word = word0;
+    inlist = inlist0;
+    while (*word && *inlist && *word == lang->to_upper(*inlist)) ++word, ++inlist;
+    if (*inlist) goto fail;
+    if (lang->special(*word).end) ++word;
+    if (*word) goto fail;
+    return true;
+  fail:
+    if (lang->special(*word0).begin) {++word0; goto try_again;}
+    return false;
   }
 
   PosibErr<void> check_if_valid(const Language & l, ParmString word) {
@@ -531,7 +467,7 @@ namespace aspeller {
 	if (lang.is_alpha(c) || lang.special(c).any)
 	  chars_set[static_cast<unsigned char>(lang.to_stripped(c))] = true;
     }
-    for (int i = 0; i != 256; ++i) 
+    for (int i = 1; i != 256; ++i) 
     {
       if (chars_set[i]) 
 	chars_list += static_cast<char>(i);
@@ -545,13 +481,14 @@ namespace aspeller {
     for (int i = 0; i != 256; ++i) 
     {
       char c = static_cast<char>(i);
-	if (lang.is_alpha(c) || lang.special(c).any)
-	  chars_set[static_cast<unsigned char>(lang.to_clean(c))] = true;
+      if (lang.is_alpha(c) || lang.special(c).any) 
+        chars_set[static_cast<unsigned char>(lang.to_clean(c))] = true;
     }
-    for (int i = 0; i != 256; ++i) 
+    for (int i = 1; i != 256; ++i) 
     {
-      if (chars_set[i]) 
+      if (chars_set[i]) {
 	chars_list += static_cast<char>(i);
+      }
     }
     return chars_list;
   }
