@@ -14,12 +14,19 @@
 
 // data block laid out as follows:
 //
-// WITH Soundslike data:
-//   (<pad><16 bit: offset to next item><8 bit: soundslike size><soundslike>
-//      (<8 bit: flags><8 bit: word size><word><null>)+<null>)+
-// WITHOUT Soundslike data
-//   (<8 bit: offset><8 bit: flags><8 bit: word size><word><null>[<affix info><null>])+
-// flags = WordInfo (see language.hpp)
+// The lower 4 bits are used for word info.  The upper 4 bits can be used to store 
+//   info on additional info.
+// * have affix flags
+// * have compound category flags
+// Words:
+//   (<8 bit frequency><8 bit: flags><8 bit: offset>
+//      <8 bit: word size><word><null>
+//      [<affix info><null>][<category info><null>])+
+// Words with soundslike:
+//   (<8 bit: offset to next item><8 bit: soundslike size><soundslike>
+//      <words>)+
+// An extra <null> is inserted at the end of each "group".
+//
 
 #include <map>
 
@@ -44,6 +51,7 @@ using std::pair;
 #include "vector.hpp"
 #include "vector_hash-t.hpp"
 #include "check_list.hpp"
+#include "lsort.hpp"
 #include "iostream.hpp"
 
 typedef unsigned int   u32int;
@@ -112,43 +120,16 @@ static inline void mmap_free(char *, unsigned int)
 
 #endif
 
-static const int NEXT_WORD_O = 3;
-static const int NEXT_SL_O = 3;
-static const int FLAGS_O = 2;
+static byte HAVE_AFFIX_FLAGS = 1 << 7;
+static byte HAVE_CATEGORY_FLAGS = 1 << 6;
+
+static byte WORD_INFO_MASK = 0x0F;
+
+static const int FREQUENCY_INFO_O = 4;
+static const int FLAGS_O = 3;
+static const int NEXT_O = 2;
 static const int WORD_SIZE_O = 1;
 
-// only use when there is no soundslike data
-static inline const char * get_next_word(const char * d) {
-  return d + *reinterpret_cast<const byte *>(d - NEXT_WORD_O);
-}
-
-static inline const char * get_next_sl(const char * d) {
-  return d + *reinterpret_cast<const u16int *>(d - NEXT_SL_O);
-}
-
-static inline const char * get_sl_words_begin(const char * d) {
-  return d + *reinterpret_cast<const byte *>(d - WORD_SIZE_O) + 3;
-}
-
-// get_next_word_for_sl might go past the end so don't JUST compare
-// for equality.  Ie use while (cur < end) not (cur != end)
-static inline const char * get_sl_words_end(const char * d) {
-  return d + *reinterpret_cast<const u16int *>(d - NEXT_SL_O) - 3;
-}
-
-static inline const char * get_next_word_for_sl(const char * d) {  
-  return d + *reinterpret_cast<const byte *>(d - WORD_SIZE_O) + 3;
-}
-
-static inline const char * get_affix(const char * d) {
-  int word_size = *reinterpret_cast<const byte *>(d - WORD_SIZE_O);
-  int affix_size = *reinterpret_cast<const byte *>(d - NEXT_WORD_O) - word_size - 5;
-  if (affix_size > 0) 
-    return d + word_size +  1;
-  else
-    return d + word_size;
-}
-
 static inline int get_word_size(const char * d) {
   return *reinterpret_cast<const byte *>(d - WORD_SIZE_O);
 }
@@ -157,9 +138,41 @@ static inline byte get_flags(const char * d) {
   return *reinterpret_cast<const byte *>(d - FLAGS_O);
 }
 
-// if affix_size <= 0 than there is no affix data
-static inline int get_affix_size(const char * d) {
-  return *reinterpret_cast<const byte *>(d - NEXT_WORD_O) - get_word_size(d) - 5;
+static inline byte get_offset(const char * d) {
+  return *reinterpret_cast<const byte *>(d - NEXT_O);
+}
+
+static inline const char * get_next(const char * d) {
+  return d + *reinterpret_cast<const byte *>(d - NEXT_O);
+}
+
+static inline const char * get_sl_words_begin(const char * d) {
+  return d + *reinterpret_cast<const byte *>(d - WORD_SIZE_O) + 4;
+  // FIXME: This isn't right when frequency info is stored in the table
+}
+
+// get_next might go past the end so don't JUST compare
+// for equality.  Ie use while (cur < end) not (cur != end)
+static inline const char * get_sl_words_end(const char * d) {
+  return get_next(d) - 3;
+}
+
+static inline const char * get_affix(const char * d) {
+  int word_size = get_word_size(d);
+  if (get_flags(d) & HAVE_AFFIX_FLAGS) 
+    return d + word_size +  1;
+  else
+    return d + word_size;
+}
+
+static inline const char * get_category(const char * d) {
+  int word_size = get_word_size(d);
+  if (get_flags(d) & HAVE_AFFIX_FLAGS & HAVE_CATEGORY_FLAGS) 
+    return d + strlen(d + word_size +  1) + 1;
+  else if (get_flags(d) & HAVE_CATEGORY_FLAGS)
+    return d + word_size +  1;
+  else
+    return d + word_size;
 }
 
 namespace {
@@ -171,11 +184,11 @@ namespace {
   //  ReadOnlyDict
   //
     
-  struct SoundslikeJump
+  struct Jump
   {
     char   sl[4];
     u32int loc;
-    SoundslikeJump() {memset(this, 0, sizeof(SoundslikeJump));}
+    Jump() : sl(), loc() {}
   };
   
   class ReadOnlyDict : public Dictionary
@@ -186,8 +199,6 @@ namespace {
     struct WordLookupParms {
       const char * block_begin;
       WordLookupParms() {}
-      //WordLookupParms(const char * b, const Language * l)
-      //	: block_begin(b), hash(l), equal(l) {}
       typedef BlockVector<const u32int> Vector;
       typedef u32int                    Value;
       typedef const char *              Key;
@@ -206,8 +217,8 @@ namespace {
     char *           block;
     u32int           block_size;
     bool             block_mmaped;
-    const SoundslikeJump * jump1;
-    const SoundslikeJump * jump2;
+    const Jump * jump1;
+    const Jump * jump2;
     WordLookup       word_lookup;
     const char *     word_block;
     
@@ -223,12 +234,6 @@ namespace {
     Size      size()     const;
     bool      empty()    const;
 
-    void convert(const char * w, WordEntry & o) const {
-      o.what = WordEntry::Word;
-      o.word = w;
-      o.aff  = !have_soundslike ? get_affix(w) :  "";
-    }
-    
     ReadOnlyDict() 
       : Dictionary(basic_dict, "ReadOnlyDict")
     {
@@ -256,6 +261,14 @@ namespace {
     
     SoundslikeEnumeration * soundslike_elements() const;
   };
+
+  static inline void convert(const char * w, WordEntry & o) {
+    o.what = WordEntry::Word;
+    o.word = w;
+    o.aff  = get_affix(w);
+    o.word_size = get_word_size(w);
+    o.word_info = get_flags(w) & WORD_INFO_MASK;
+  }
     
   //
   //  
@@ -270,7 +283,7 @@ namespace {
     bool endf(const Iterator & i) const {return i.at_end();}
     Value end_state() const {return 0;}
     WordEntry * deref(const Iterator & i) {
-      ws->convert(ws->word_block + *i, data);
+      convert(ws->word_block + *i, data);
       return & data;
     }
   };
@@ -287,7 +300,7 @@ namespace {
     return word_lookup.empty();
   }
 
-  static const char * const cur_check_word = "aspell default speller rowl 1.8b";
+  static const char * const cur_check_word = "aspell default speller rowl 1.9";
 
   struct DataHead {
     // all sizes except the last four must to divisible by:
@@ -312,7 +325,11 @@ namespace {
     u32int soundslike_name_size;
     u32int soundslike_version_size;
 
-    u32int affix_info; // none zero if affix information is encoded in table
+    byte affix_info; // 0 = none, 1 = partially expanded, 2 = full
+    byte invisible_soundslike;
+    byte soundslike_root_only;
+    byte compound_info; //
+    byte freq_info;
   };
 
   PosibErr<void> ReadOnlyDict::load(ParmString f0, Config & config, 
@@ -359,8 +376,6 @@ namespace {
     }
 
     if (data_head.soundslike_name_size != 0) {
-      have_soundslike = true;
-
       word.resize(data_head.soundslike_name_size);
       f.read(word.data(), data_head.soundslike_name_size);
 
@@ -374,6 +389,9 @@ namespace {
         return make_err(bad_file_format, fn, _("Wrong soundslike version."));
     }
 
+    invisible_soundslike = data_head.invisible_soundslike;
+    soundslike_root_only = data_head.soundslike_root_only;
+
     affix_compressed = data_head.affix_info;
 
     block_size = data_head.block_size;
@@ -387,15 +405,13 @@ namespace {
 
     if (data_head.jump2_offset) {
       fast_scan = true;
-      jump1 = reinterpret_cast<const SoundslikeJump *>(block 
-						       + data_head.jump1_offset);
-      jump2 = reinterpret_cast<const SoundslikeJump *>(block 
-						       + data_head.jump2_offset);
+      jump1 = reinterpret_cast<const Jump *>(block + data_head.jump1_offset);
+      jump2 = reinterpret_cast<const Jump *>(block + data_head.jump2_offset);
     } else {
       jump1 = jump2 = 0;
     }
 
-    word_block       = block + data_head.word_offset;
+    word_block = block + data_head.word_offset;
 
     word_lookup.parms().block_begin = word_block;
     word_lookup.parms().hash .lang     = lang();
@@ -428,29 +444,24 @@ namespace {
   {
     WordEntry data;
     const ReadOnlyDict * obj;
-    const SoundslikeJump * jump1;
-    const SoundslikeJump * jump2;
+    const Jump * jump1;
+    const Jump * jump2;
     const char * cur;
     int level;
-    bool sl;
-
-    const char * next_sl() const {
-      if (sl) return get_next_sl(cur);
-      else return get_next_word(cur);
-    }
+    bool invisible_soundslike;
 
     WordEntry * next(int stopped_at);
 
     SoundslikeElements(const ReadOnlyDict * o)
       : obj(o), jump1(obj->jump1), jump2(obj->jump2), cur(0), 
-        level(1), sl(o->have_soundslike) {
-      data.what = o->have_soundslike ? WordEntry::Soundslike : WordEntry::Word;}
+        level(1), invisible_soundslike(o->invisible_soundslike) {
+      data.what = o->invisible_soundslike ? WordEntry::Word : WordEntry::Soundslike;}
   };
 
   WordEntry * ReadOnlyDict::SoundslikeElements::next(int stopped_at) {
 
     //CERR << level << ":" << stopped_at << "  :";
-    //CERR << jump1->sl << ":" << jump2->sl << "\n";
+    //CERR << jump1->data << ":" << jump2->sl << "\n";
 
     const char * tmp = cur;
 
@@ -482,10 +493,9 @@ namespace {
     } else if (level == 2) {
 
       cur = tmp = obj->word_block + jump2->loc;
-      cur = next_sl(); // next pos uses cur
       level = 3;
 
-    } else if (next_sl() == cur) {
+    } else if (get_offset(cur) == 0) {
 
       level = 2;
       ++jump2;
@@ -498,25 +508,25 @@ namespace {
       }
       goto jquit;
 
-    } else {
+    } 
 
-      cur = next_sl();
-      
-    }
+    cur = get_next(cur); // this will be the NEXT item looked at
 
     data.word = tmp;
-    if (!sl) {
+    data.word_size = get_word_size(tmp);
+    if (invisible_soundslike) {
+      convert(tmp, data);
       data.what = WordEntry::Word;
-      data.aff  = get_affix(tmp);
-    }
+    } 
     data.intr[0] = (void *)tmp;
     return &data;
 
   jquit:
     if (!*tmp) return 0;
     data.word = tmp;
+    data.word_size = !tmp[1] ? 1 : !tmp[2] ? 2 : 3;
     data.intr[0] = 0;
-    if (!sl) {
+    if (invisible_soundslike) {
       data.what = WordEntry::Clean;
       data.aff  = 0;
     }
@@ -537,11 +547,10 @@ namespace {
   WordEntry * ReadOnlyDict::CleanElements::next(int) {
 
     const char * tmp = cur;
-    cur = get_next_word(cur);
+    cur = get_next(cur);
     if (cur == tmp) return 0;
     data.intr[0] = (void *)tmp;
-    data.word = tmp;
-    data.aff  = get_affix(tmp);
+    convert(tmp, data);
     return &data;
 
   }
@@ -559,8 +568,8 @@ namespace {
   {
     const char * cur = (const char *)(w->intr[0]);
     const char * end = (const char *)(w->intr[1]);
-    w->word = cur;
-    cur = get_next_word_for_sl(cur);
+    convert(cur, *w);
+    cur = get_next(cur);
     w->intr[0] = (void *)cur;
     if (cur >= end) w->adv_ = 0;
   }
@@ -585,7 +594,7 @@ namespace {
 
       return false;
 
-    } else if (have_soundslike) {
+    } else if (!invisible_soundslike) {
       
       w.clear();
       w.what = WordEntry::Word;
@@ -607,7 +616,7 @@ namespace {
 
   bool ReadOnlyDict::soundslike_lookup(ParmString s, WordEntry & w) const 
   {
-    if (!have_soundslike) {
+    if (invisible_soundslike) {
       return ReadOnlyDict::clean_lookup(s,w);
     } else {
       return false;
@@ -626,73 +635,55 @@ namespace aspeller {
 
 namespace {
 
+  // Possible:
+  //   No Affix Compression:
+  //     no soundslike
+  //     invisible soundslike
+  //     with soundslike
+  //   Affix Compression:
+  //     group by root:
+  //       no soundslike
+  //       invisible soundslike
+  //       with soundslike
+  //     expand prefix:  
+  //       no soundslike
+  //       invisible soundslike
+
   using namespace aspeller;
 
-  // TODO:
-  //   Store soundslike with word and eliminate map
-  // struct WordData 
-  //   WordData * sl_next;
-  //   char total_size; // including null chars
-  //   char word_size; // not including null char
-  //   char sl_size; // start at total_size
-  //   char data[]; <word> <affix> <soundslike>
-  // then sort soundslike linked-list  
-
   struct WordData {
-    static const unsigned int size = 2;
-    unsigned char total_size; // including null chars
-    unsigned char word_size; // not including null char
-    char data[];
-    const char * affix_data() const {
-      return total_size == word_size + 1 ? 0 : data + word_size + 1;
+    WordData * next;
+    char * sl;
+    byte word_size;
+    byte sl_size;
+    byte data_size;
+    byte flags;
+    char word[];
+  };
+
+  struct SoundslikeLess {
+    bool operator() (WordData * x, WordData * y) const {
+      int res = strcmp(x->sl, y->sl);
+      if (res != 0) return res < 0;
+      return strcmp(x->word, y->word) < 0;
     }
   };
-  
+
   struct WordLookupParms {
-    typedef Vector<const WordData *> Vector;
-    typedef const WordData *         Value;
-    typedef const char *             Key;
+    const char * block_begin;
+    WordLookupParms() {}
+    typedef Vector<u32int>      Vector;
+    typedef u32int              Value;
+    typedef const char *        Key;
     static const bool is_multi = true;
-    Key key(Value v) const {return v->data;}
-    InsensitiveHash hash_;
-    size_t hash(Key k) const {return hash_(k);}
-    InsensitiveEqual equal_;
-    bool equal(Key rhs, Key lhs) const {
-      return equal_(rhs, lhs);
-    }
-    bool is_nonexistent(const Value & v) const {return v == 0;}
-    void make_nonexistent(Value & v) const {v = 0;}
+    Key key(Value v) const {assert (v != u32int_max);
+                            return block_begin + v;}
+    InsensitiveHash  hash;
+    InsensitiveEqual equal;
+    bool is_nonexistent(Value v) const {return v == u32int_max;}
+    void make_nonexistent(Value & v) const {v = u32int_max;}
   };
-
-  typedef VectorHashTable<WordLookupParms> WordHash;
-  
-  struct SoundslikeList {
-    struct Word {
-      const WordData * d;
-      u32int hash_idx;
-    };
-    union {
-      Word * list;
-      Word   single;
-    } d;
-    u16int   size;
-    u16int   num_inserted;
-  };
-
-  struct SoundslikeLt
-  {
-    InsensitiveCompare cmp;
-    SoundslikeLt(const Language * l = 0)
-      : cmp(l) {}
-    bool operator()(const char * s1, const char * s2) const {
-      if (cmp)
-        return cmp(s1, s2) < 0;
-      else
-        return strcmp(s1, s2) < 0;
-    }
-  };
-
-  typedef std::multimap<const char *, SoundslikeList, SoundslikeLt> SoundMap;
+  typedef VectorHashTable<WordLookupParms> WordLookup;
 
   static inline unsigned int round_up(unsigned int i, unsigned int size) {
     return ((i + size - 1)/size)*size;
@@ -712,13 +703,24 @@ namespace {
     assert(sizeof(u16int) == 2);
     assert(sizeof(u32int) == 4);
 
+    bool full_soundslike = (strcmp(lang.soundslike_name(), "none") != 0 && 
+                            strcmp(lang.soundslike_name(), "simple"));
+
     bool affix_compress = (lang.affix() && 
                            config.retrieve_bool("affix-compress"));
-    
-    bool use_soundslike = (strcmp(lang.soundslike_name(),"none") != 0 &&
-                           config.retrieve_bool("use-soundslike"));
-   
-    bool use_jump_tables = use_soundslike || config.retrieve_bool("use-jump-tables");
+
+    bool partially_expand = (affix_compress &&
+                             !full_soundslike &&
+                             config.retrieve_bool("partially-expand"));
+
+    bool invisible_soundslike = false;
+    if (partially_expand)
+      invisible_soundslike = true;
+    else if (config.have("invisible-soundslike"))
+      invisible_soundslike = config.retrieve_bool("invisible-soundslike");
+    else if (!full_soundslike)
+      invisible_soundslike = true;
+
 
     ConvEC iconv;
     if (!config.have("norm-strict"))
@@ -728,15 +730,6 @@ namespace {
     else
       RET_ON_ERR(iconv.setup(config, lang.data_encoding(), lang.charset(),NormFrom));
 
-    //CERR << (affix_compress ? "  AFFIX COMPRESS" : "")
-    //     << (use_soundslike ? "  USING SOUNDSLIKE" : "") 
-    //     << (use_jump_tables ? "  USING JUMP TABLES" : "")
-    //     << "\n";
-
-    if (affix_compress && use_soundslike)
-      return make_err(other_error, 
-                      _("Affix compression is currently incompatible with soundslike lookup."));
-
     String base = config.retrieve("master-path");
 
     DataHead data_head;
@@ -747,10 +740,8 @@ namespace {
 
     data_head.dict_name_size = 1;
     data_head.lang_name_size = strlen(lang.name()) + 1;
-    if (use_soundslike) {
-      data_head.soundslike_name_size    = strlen(lang.soundslike_name()) + 1;
-      data_head.soundslike_version_size = strlen(lang.soundslike_version()) + 1;
-    }
+    data_head.soundslike_name_size    = strlen(lang.soundslike_name()) + 1;
+    data_head.soundslike_version_size = strlen(lang.soundslike_version()) + 1;
     data_head.head_size  = sizeof(DataHead);
     data_head.head_size += data_head.dict_name_size;
     data_head.head_size += data_head.lang_name_size;
@@ -758,41 +749,40 @@ namespace {
     data_head.head_size += data_head.soundslike_version_size;
     data_head.head_size  = round_up(data_head.head_size, DataHead::align);
 
-    data_head.affix_info = affix_compress;
+    data_head.affix_info = affix_compress ? partially_expand ? 1 : 2 : 0;
+    data_head.invisible_soundslike = invisible_soundslike;
+    data_head.soundslike_root_only = affix_compress  && !partially_expand ? 1 : 0;
 
+//#if 0
+    CERR.printl("FLAGS:  ");
+    if (full_soundslike) CERR.printl("  full soundslike");
+    if (invisible_soundslike) CERR.printl("  invisible soundslike");
+    if (data_head.soundslike_root_only) CERR.printl("  soundslike root only");
+    if (affix_compress) CERR.printl("  affix compress");
+    if (partially_expand) CERR.printl("  partially expand");
+    CERR.printl("---");
+//#endif
+    
     String temp;
 
-    // these are the final data structures
-    CharVector     data;
-    Vector<u32int> final_hash;
-    Vector<SoundslikeJump> jump1;
-    Vector<SoundslikeJump> jump2;
-
     int num_entries = 0;
     
-    StringBuffer   buf;
-    SoundslikeLt   soundslike_lt;
-    if (!use_soundslike)
-      soundslike_lt.cmp.lang = &lang;
-    SoundMap       sound_map(soundslike_lt);
-    Vector<SoundslikeList::Word> sl_list_buf;
-    int                          sl_list_buf_size = 0;
+    ObjStack buf(16*1024);
+    String sl_buf;
 
-    StackPtr<WordHash> word_hash(new WordHash);
+    WordData * first = 0;
 
     //
-    // Read in Wordlist from stdin and create initial Word Hash
+    // Read in Wordlist
     //
     {
-      word_hash->parms().hash_ .lang = &lang;
-      word_hash->parms().equal_.cmp.lang = &lang;
       const char * w0;
-      WordHash::MutableFindIterator j;
       ObjStack exp_buf;
       WordAff * exp_list;
       WordAff single;
       single.next = 0;
       Vector<WordAff> af_list;
+      WordData * * prev = &first;
 
       while ( (w0 = els->next()) != 0) {
 
@@ -822,330 +812,261 @@ namespace {
           return make_err(other_error, 
                           _("Affix flags found in word but no affix file given."));
 
-        if (!affixes || (affixes && !affix_compress)) {
-
-          // Now expand any affix compression
-          
-          if (affixes) {
-            exp_buf.reset();
-            exp_list = lang.affix()->expand(w, affixes, exp_buf);
-          } else {
-            exp_list = &single;
-            single.word.str = w;
-            single.word.size = strlen(w);
-          }
-
-          // iterate through each expanded word
-          
-          for (WordAff * p = exp_list; p; p = p->next)
-          {
-            const char * w = p->word.str;
-            s = p->word.size;
-          
-            bool dup = false;
-            for (j =word_hash->multi_find(w); !j.at_end(); j.adv()) {
-              if (strcmp(w, j.deref()->data)==0) {
-                dup = true;
-              }
-            }
-            if (dup) continue;
-          
-            WordData * b = (WordData *)buf.alloc(s + WordData::size + 1);
-            b->total_size = s + 1;
-            b->word_size = s;
-            memcpy(b->data, w, s+1);
-            word_hash->insert(b);
-            ++num_entries;
-          }
-
+        if (affixes && !affix_compress) {
+          exp_buf.reset();
+          exp_list = lang.affix()->expand(w, affixes, exp_buf);
+        } else if (affixes && partially_expand) {
+          // expand any affixes which will effect the first
+          // 3 letters of a word.  This is needed so that the
+          // jump tables will function correctly
+          exp_buf.reset();
+          exp_list = lang.affix()->expand(w, affixes, exp_buf, 3);
         } else {
+          single.word.str = w;
+          single.word.size = strlen(w);
+          single.aff = (const byte *)(affixes ? affixes : "");
+          exp_list = &single;
+        }
 
-          if (use_jump_tables) {
-            // expand any affixes which will effect the first
-            // 3 letters of a word.  This is needed so that the
-            // jump tables will function correctly
-            exp_buf.reset();
-            exp_list = lang.affix()->expand(w, affixes, exp_buf, 3);
+        // iterate through each expanded word
+        
+        for (WordAff * p = exp_list; p; p = p->next)
+        {
+          const char * w = p->word.str;
+          s = p->word.size;
+          
+          unsigned total_size = sizeof(WordData);
+          unsigned data_size = s + 1;
+          unsigned aff_size = strlen((const char *)p->aff);
+          if (aff_size > 0) data_size += aff_size + 1;
+          total_size += data_size;
+          lang.to_soundslike(sl_buf, w);
+          const char * sl = sl_buf.str();
+          unsigned sl_size = sl_buf.size();
+          if (strcmp(sl,w) == 0) sl = w;
+          if (sl != w) total_size += sl_size + 1;
+
+          if (total_size - sizeof(WordData) > 240)
+            return make_err(invalid_word, MsgConv(lang)(w),
+                            _("The total word length, with soundslike data, is larger than 240 characters."));
+
+          WordData * b = (WordData *)buf.alloc(total_size);
+          *prev = b;
+          b->next = 0;
+          prev = &b->next;
+          
+          b->word_size = s;
+          b->sl_size = strlen(sl);
+          b->data_size = data_size;
+          b->flags = lang.get_word_info(w);
+
+          char * z = b->word;
+
+          memcpy(z, w, s + 1);
+          z += s + 1;
+
+          if (aff_size > 0) {
+            b->flags |= HAVE_AFFIX_FLAGS;
+            memcpy(z, p->aff, aff_size + 1);
+            z += aff_size + 1;
+          }
+
+          if (sl != w) {
+            memcpy(z, sl, sl_size + 1);
+            b->sl = z;
           } else {
-            single.word.str = w;
-            single.word.size = strlen(w);
-            single.aff = (const byte *)affixes;
-            exp_list = &single;
+            b->sl = b->word;
           }
 
-          // now interate through the list and insert
-          // the words 
-          for (WordAff * p = exp_list; p; p = p->next)
-          {
-            //CERR.printf("%s|%s\n", p->word.str, p->aff);
-            s = p->word.size;
-            const char * w = p->word.str;
-            bool dup = false;
-            for (j =word_hash->multi_find(w); !j.at_end(); j.adv()) {
-              if (strcmp(w, j.deref()->data)==0) {
-                if (p->aff)
-                  CERR << "WARNING: Ignoring duplicate: " << MsgConv(lang)(w) << '\n';
-                dup = true;
-                // FIXME: deal with duplicates properly
-              }
-            }
-            if (dup) continue;
-
-            WordData * b;
-            size_t aff_size = strlen((const char *)p->aff);
-            b = (WordData *)buf.alloc(s + WordData::size + aff_size + 2);
-            b->total_size = s + 1 + aff_size + 1;
-            b->word_size = s;
-
-            memcpy(b->data, w, s + 1);
-            memcpy(b->data + s + 1, p->aff, aff_size + 1);
-
-            word_hash->insert(b);
-            ++num_entries;
-          }
+          ++num_entries;
         }
       }
       delete els;
     }
 
-    word_hash->resize(word_hash->size()*4/5);
+    //
+    // sort WordData linked list based on (sl, word)
+    //
+
+    first = sort(first, SoundslikeLess());
 
     //
+    // duplicate checks
     // 
-    //
-    if (use_soundslike) {
-
-      for (unsigned int i = 0; i != word_hash->vector().size(); ++i) {
-        
-        const WordData * value = word_hash->vector()[i];
-        
-        if (word_hash->parms().is_nonexistent(value)) continue;
-        
-        lang.LangImpl::to_soundslike(temp, value->data);
-        SoundMap::iterator j = sound_map.find(temp.str());
-        if (j == sound_map.end()) {
-          
-          SoundMap::value_type to_insert(buf.dup(temp), SoundslikeList());
-          sound_map.insert(to_insert)->second.size = 1;
-          
-        } else {
-          
-          if (j->second.size == 1)
-            sl_list_buf_size++;
-          
-          j->second.size++;
-          sl_list_buf_size++;
-	  
-        }
-        
+    WordData * prev = first;
+    WordData * cur = first->next;
+    while (cur) {
+      if (strcmp(prev->word, cur->word) == 0) {
+        CERR.printf("WARNING: Ignoring duplicate: %s\n", cur->word);
+        prev->next = cur->next;
+      } else {
+        prev = cur;
       }
-
-      sl_list_buf.resize(sl_list_buf_size);
-      int p = 0;
-      
-      for (unsigned int i = 0; i != word_hash->vector().size(); ++i) {
-	
-	const WordData * value = word_hash->vector()[i];
-	
-	if (word_hash->parms().is_nonexistent(value)) continue;
-
-        SoundMap::iterator j;
-        lang.LangImpl::to_soundslike(temp.data(), value->data);
-        j = sound_map.find(temp.data());
-        assert(j != sound_map.end());
-	if (j->second.num_inserted == 0 && j->second.size != 1) {
-	  j->second.d.list = &*sl_list_buf.begin() + p;
-	  p += j->second.size;
-	} 
-	if (j->second.size == 1) {
-	  j->second.d.single.d        = value;
-	  j->second.d.single.hash_idx = i;
-	} else {
-	  j->second.d.list[j->second.num_inserted].d        = value;
-	  j->second.d.list[j->second.num_inserted].hash_idx = i;
-	}
-	++j->second.num_inserted;
-      }
-
-      data_head.soundslike_count = sound_map.size();
-      
-    } else if (use_jump_tables) {
-      
-      for (unsigned int i = 0; i != word_hash->vector().size(); ++i) {
-
-        const WordData * value = word_hash->vector()[i];
-        
-        if (word_hash->parms().is_nonexistent(value)) continue;
-
-        SoundMap::value_type to_insert(value->data, 
-                                       SoundslikeList());
-        SoundslikeList & sl = sound_map.insert(to_insert)->second;
-        sl.size = 1;
-        sl.d.single.d        = value;
-        sl.d.single.hash_idx = i;
-        sl.num_inserted      = 1;
-      }
-
+      cur = cur->next;
     }
 
-    data_head.word_count   = word_hash->size();
-    data_head.word_buckets = word_hash->bucket_count();
-
-    final_hash.insert(final_hash.begin(), 
-		      word_hash->bucket_count(), u32int_max);
-
-    
     //
-    // Create the data block which includes both the word and soundlike
-    //  data
+    //
     //
 
-    if (use_jump_tables) {
-
-      word_hash.del(); // the word_hash is no longer needed so free the space
-
-      SoundMap::iterator i   = sound_map.begin();
-      SoundMap::iterator end = sound_map.end();
-
-      data.write32(0); // to avoid nasty special cases
-      unsigned int prev_pos = data.size();
-      data.write32(0);
-      String sl;
-      String prev_sl = "";
-
-      for (; i != end; ++i) {
-
-	if (use_soundslike) {
-
-	  sl = i->first;
-          if (data.size() % 2 != 0) data.write('\0'); // align
-          data.write16(0); // place holder for offset to next item
-	  data.write((byte)sl.size());
-
-	} else {
-
-	  lang.LangImpl::to_clean(sl, i->first);
-          data.write('\0'); // place holder for offset to next item
-          data.write('\0'); // flags  
-          data.write('\0'); // word size
-
-	}
-
-	if (strncmp(prev_sl.c_str(), sl.c_str(), 3) != 0) {
-
-	  SoundslikeJump jump;
-	  strncpy(jump.sl, sl.c_str(), 3);
-	  jump.loc = data.size();
-	  jump2.push_back(jump);
-
-	  if (strncmp(prev_sl.c_str(), sl.c_str(), 2) != 0) {
-	    SoundslikeJump jump;
-	    strncpy(jump.sl, sl.c_str(), 2);
-	    jump.loc = jump2.size() - 1;
-	    jump1.push_back(jump);
-	  }
-
-          if (use_soundslike)
-            data.at16(prev_pos - NEXT_SL_O) = 0;
-          else
-            data[prev_pos - NEXT_WORD_O] = 0;
-
-	} else {
-
-          if (use_soundslike)
-            data.at16(prev_pos - NEXT_SL_O) = data.size() - prev_pos;
-          else
-            data[prev_pos - NEXT_WORD_O] = (byte)(data.size() - prev_pos);
-
-	}
-
-	prev_pos = data.size();
-
-	// Write soundslike
-	if (use_soundslike)
-	  data << sl << '\0';
-
-	// Write words
-	{
-	  SoundslikeList::Word * k 
-	    = i->second.size == 1 ? &(i->second.d.single) : i->second.d.list;
-	  SoundslikeList::Word * end = k + i->second.size;
-          assert(use_soundslike || i->second.size == 1);
-	  for (;k != end; ++k) {
-            if (use_soundslike) {
-              data.write((byte)lang.get_word_info(k->d->data));
-              data.write((byte)k->d->word_size);
-            } else {
-              data[prev_pos - WORD_SIZE_O] = k->d->word_size;
-            }
-
-	    final_hash[k->hash_idx] = data.size();
-
-            data << k->d->data << '\0';
-
-            if (k->d->affix_data())
-              data << k->d->affix_data() << '\0';
-	  }
-	}
-
-	prev_sl = sl;
-      }
-
-      // add special end case
-      if (data.size() % 2 != 0) data.write('\0');
-      data.write16(0);
-      data.write16(0);
-      if (use_soundslike)
-        data.at16(prev_pos - NEXT_SL_O) |= data.size() - prev_pos;
-      else
-        data[prev_pos - NEXT_WORD_O] |= (byte)(data.size() - prev_pos);
-      data.at32(4) = 0;
-
-      jump2.push_back(SoundslikeJump());
-      jump1.push_back(SoundslikeJump());
+    unsigned data_size = 16;
+    WordData * p = first;
+    if (invisible_soundslike) {
+      
+      for (; p; p = p->next)
+        data_size += 3 + p->data_size;
 
     } else {
 
-      // No jump tables, just write the words in the order that
-      // they appear in the hash table.
+      while (p)
+      {
+        unsigned ds = 2 + p->sl_size + 1;
 
-      for (unsigned int i = 0; i != word_hash->vector().size(); ++i) {
+        char * prev = p->sl;
 
-        const WordData * value = word_hash->vector()[i];
-        
-        if (word_hash->parms().is_nonexistent(value)) continue;
+        do {
+          
+          ds += 3 + p->data_size;
+          p->sl = prev;
+          p = p->next;
 
-        data.write(value->total_size + 3);
-        data.write((byte)lang.get_word_info(value->data));
-        data.write(value->word_size);
+        } while (p && strcmp(prev, p->sl) == 0 && ds + 3 + p->data_size < 255);
 
-        final_hash[i] = data.size();
+        data_size += ds;
 
-        data.write(value->data, value->total_size);
-        
       }
 
-      data.write(0);
-      data.write(0);
-      data.write(0);
-
     }
 
+    //
+    // Create the final data structures
+    //
+
+    CharVector     data;
+    data.reserve(data_size);
+    data.write32(0); // to avoid nasty special cases
+    unsigned int prev_pos = data.size();
+    data.write32(0);
+
+    WordLookup lookup(num_entries * 4 / 5);
+    lookup.parms().block_begin = data.begin();
+    lookup.parms().hash .lang     = &lang;
+    lookup.parms().equal.cmp.lang = &lang;
+
+    Vector<Jump> jump1;
+    Vector<Jump> jump2;
+
+    const int head_size = invisible_soundslike ? 3 : 2;
+
+    const char * prev_sl = "";
+    p = first;
+    while (p)
+    {
+      if (invisible_soundslike) {
+
+        data.write(p->flags); // flags  
+        data.write('\0'); // place holder for offset to next item
+        data.write(p->word_size);
+
+      } else {
+
+        data.write('\0'); // place holder for offset to next item
+        data.write(p->sl_size);
+
+      }
+        
+      if (strncmp(prev_sl, p->sl, 3) != 0) {
+        
+        Jump jump;
+        strncpy(jump.sl, p->sl, 3);
+        jump.loc = data.size();
+        jump2.push_back(jump);
+        
+        if (strncmp(prev_sl, p->sl, 2) != 0) {
+          Jump jump;
+          strncpy(jump.sl, p->sl, 2);
+          jump.loc = jump2.size() - 1;
+          jump1.push_back(jump);
+        }
+
+        data[prev_pos - NEXT_O] = (byte)(data.size() - prev_pos - head_size + 1);
+        // when advanced to this position the offset byte will
+        // be null (since it will point to the null terminator
+        // of the last word) and will thus signal the end of the
+        // group
+        
+      } else {
+        
+        data[prev_pos - NEXT_O] = (byte)(data.size() - prev_pos);
+        
+      }
+        
+      prev_pos = data.size();
+      prev_sl = p->sl;
+
+      if (invisible_soundslike) {
+        
+        unsigned pos = data.size();
+        data.write(p->word, p->data_size);
+        lookup.insert(pos);
+
+        p = p->next;
+
+      } else {
+
+        data.write(p->sl, p->sl_size + 1);
+
+        // write all word entries with the same soundslike
+
+        const char * prev;
+
+        do {
+          data.write(p->flags);
+          data.write(p->data_size + 3);
+          data.write(p->word_size);
+
+          unsigned pos = data.size();
+          data.write(p->word, p->data_size);
+          lookup.insert(pos);
+
+          prev = p->sl;
+
+          p = p->next;
+
+        } while (p && prev == p->sl); // yes I really mean to use pointer compare here
+      }
+    }
+
+    // add special end case
+    if (data.size() % 2 != 0) data.write('\0');
+    data.write16(0);
+    data.write16(0);
+    data[prev_pos - NEXT_O] |= (byte)(data.size() - prev_pos);
+    data.at32(4) = 0;
+    
+    jump2.push_back(Jump());
+    jump1.push_back(Jump());
+    
+    data.write(0);
+    data.write(0);
+    data.write(0);
+
+    data_head.word_count   = lookup.size();
+    data_head.word_buckets = lookup.bucket_count();
+
     FStream out;
     out.open(base, "wb");
 
     advance_file(out, data_head.head_size);
 
-    if (use_jump_tables) {
-      // Write jump1 table
-      data_head.jump1_offset = out.tell() - data_head.head_size;
-      out.write(jump1.data(), jump1.size() * sizeof(SoundslikeJump));
-      
-      // Write jump2 table
-      advance_file(out, round_up(out.tell(), DataHead::align));
-      data_head.jump2_offset = out.tell() - data_head.head_size;
-      out.write(jump2.data(), jump2.size() * sizeof(SoundslikeJump));
-    } 
+    // Write jump1 table
+    data_head.jump1_offset = out.tell() - data_head.head_size;
+    out.write(jump1.data(), jump1.size() * sizeof(Jump));
+    
+    // Write jump2 table
+    advance_file(out, round_up(out.tell(), DataHead::align));
+    data_head.jump2_offset = out.tell() - data_head.head_size;
+    out.write(jump2.data(), jump2.size() * sizeof(Jump));
 
     // Write data block
     advance_file(out, round_up(out.tell(), DataHead::align));
@@ -1155,7 +1076,7 @@ namespace {
     // Write hash
     advance_file(out, round_up(out.tell(), DataHead::align));
     data_head.hash_offset = out.tell() - data_head.head_size;
-    out.write(&final_hash.front(), final_hash.size() * 4);
+    out.write(&lookup.vector().front(), lookup.vector().size() * 4);
     
     // calculate block size
     advance_file(out, round_up(out.tell(), DataHead::align));
@@ -1166,10 +1087,8 @@ namespace {
     out.write(&data_head, sizeof(DataHead));
     out.write(" ", 1);
     out.write(lang.name(), data_head.lang_name_size);
-    if (use_soundslike) {
-      out.write(lang.soundslike_name(), data_head.soundslike_name_size);
-      out.write(lang.soundslike_version(), data_head.soundslike_version_size);
-    }
+    out.write(lang.soundslike_name(), data_head.soundslike_name_size);
+    out.write(lang.soundslike_version(), data_head.soundslike_version_size);
 
     return no_err;
   }
