@@ -13,6 +13,22 @@ namespace acommon {
 
 namespace aspeller {
 
+  // WordInfo
+
+  typedef unsigned int WordInfo; // 4 bits
+
+  enum CasePattern {Other, FirstUpper, AllLower, AllUpper};
+  //   Other      00
+  //   FirstUpper 01
+  //   AllLower   10
+  //   AllUpper   11
+  // First bit : is upper
+  // Second bit: uniform case
+
+  static const WordInfo CASE_PATTERN = 3;
+  static const WordInfo ALL_PLAIN    = (1 << 2);
+  static const WordInfo ALL_CLEAN    = (1 << 3);
+
   using namespace acommon;
 
   class Language;
@@ -26,35 +42,23 @@ namespace aspeller {
   {
     const char * word;
     const char * aff;
+    const char * catg;
     void (* adv_)(WordEntry *);
     void (* free_)(WordEntry *);
     void * intr[2];
+    unsigned word_size;
     enum What {Other, Word, Soundslike, Clean, Misspelled} what;
+    WordInfo word_info;
+    int frequency; // 0 .. 255
     // if type is Word than aff will be defined, otherwise it won't
     bool at_end() {return !word;}
     bool adv() {if (adv_) {adv_(this); return true;} word = 0; return false;}
     operator bool () const {return word != 0;}
-    OStream & write(OStream & o, const Language & l,
-		    const ConvertWord &, Convert * c = 0) const;
-    WordEntry() : word(0), aff(0), adv_(0), free_(0), intr() {}
-    void clear() {if (free_) free_(this); word = 0; aff = 0; adv_ = 0; free_ = 0;}
+    OStream & write(OStream & o, const Language & l, Convert * c = 0) const;
+    WordEntry() {memset(this, 0, sizeof(WordEntry));}
+    void clear() {if (free_) free_(this); memset(this, 0, sizeof(WordEntry));}
     ~WordEntry() {if (free_) free_(this);}
   };
-
-  /*
-    flags:
-      1 bit:  case/accent info known
-      1 bit:  all lower
-      1 bit:  all upper
-      1 bit:  title
-      1 bit:  with accents
-      1 bit:  special characters
-
-      4 Case Possibilities
-        All Upper, All Lower, Title, Other
-      With or without accents
-      With or without "special" characyers
-  */
 }
 
 #endif
